/*---------------------------------------------------------------------------*/
/*
   sc4i34.c

   Demonstration program for configuration of the Mesa Electronics 4I34 I/O
	 card.

   Version 1.1, Monday June 9, 2003 -- 15:21:41.
*/
/*---------------------------------------------------------------------------*/
/*
   Compiler: Borland C++, version 3.1+.

   Tabs: 4
*/
/*---------------------------------------------------------------------------*/
/*
   Revision history.

   1) Version 1.0, Thursday October 4, 2001 -- 18:08:08.

	  Code frozen for version 1.0.


   2) Version 1.1, Monday June 9, 2003 -- 15:21:41.

      Changes from version 1.0:
	  1) Added: readbytesfromfile(), saybytesfromfile(), sayfileinfo(),
		 outportbwswap(), outportbnoswap(), selectfiletype(), DataOutFuncPtr.

	  2) Removed: ValidPortList[], IOBASEADDRESSES.

      3) Modified: fatalerror(), programcard().
*/
/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#if defined __TURBOC__
/*
   Assume Borland C++/Turbo C++.

   Minimal compiler command line: "bcc sc4i34.c"
*/
#include <dos.h>
#else
/*
   (Lazy programmer assumes that we're compiling on a Linux machine with
	 kernel 2.2+ + glibc2 running on an 'x86 CPU.)

   Minimal compiler command line: "gcc -O sc4i34.c -o sc4i34"

   Note that the "-O" is required.	The resulting program must be run as root
	 Don't try this without adult supervision.
*/
#include <sys/io.h> /* (For inb(), outb(), ioperm().) */
#define inportb(p)	   inb(p)
#define outportb(p, v) outb(v, p)
#define AS_LINUX
#endif

/*---------------------------------------------------------------------------*/

/* Assorted #defines.
*/
#define PROGNAME      "SC4I34"
#define SIGNONMESSAGE "\n"PROGNAME" version "PROGVERSION"\n"
#define PROGVERSION   "1.1"


/* I/O register indices.
*/
#define R_4I34DATA    0 /* 4I34 data register. */
#define R_4I34CONTROL 1 /* 4I34 control register. */
#define R_4I34STATUS  1 /* 4I34 status register. */
#define NUM4I34REGS   2


/* Masks for R_4I34CONTROL.
*/
#define B_4I34CFGCS           0 /* Chip select. Enables read/write access to the FPGA via R_4I34DATA. */
#define M_4I34CFGCSON         (0x0 << B_4I34CFGCS) /* Enable access. */
#define M_4I34CFGCSOFF        (0x1 << B_4I34CFGCS) /* Disable access. */

#define B_4I34CFGINIT         1 /* Control programming. */
#define M_4I34CFGINITASSERT   (0x0 << B_4I34CFGINIT) /* Wipe current configuration, prepare for programming. */
#define M_4I34CFGINITDEASSERT (0x1 << B_4I34CFGINIT) /* Wait for programming data. */

#define B_4I34CFGWRITE        2 /* Data direction control for R_4I34DATA. */
#define M_4I34CFGWRITEENABLE  (0x0 << B_4I34CFGWRITE) /* CPU --> R_4I34DATA. */
#define M_4I34CFGWRITEDISABLE (0x1 << B_4I34CFGWRITE) /* R_4I34DATA --> CPU. */

#define B_4I34LED             3 /* Red LED control. */
#define M_4I34LEDON           (0x1 << B_4I34LED)
#define M_4I34LEDOFF          (0x0 << B_4I34LED)


/* Masks for R_4I34STATUS.
*/
#define B_4I34PROGDUN 0 /* Programming-done flag. Set when the FPGA "program" has been successfully uploaded to the 4I34. */
#define M_4I34PROGDUN (0x1 << B_4I34PROGDUN)


/* Other constants.
*/
#define PROGWAITLOOPCOUNT 20000 /* Wait loop count for programming completion. */


/* Exit codes.
*/
#define EC_OK    0   /* Exit OK. */
#define EC_BADCL 100 /* Bad command line. */
#define EC_USER  101 /* Invalid input. */
#define EC_HDW   102 /* Some sort of hardware failure on the 4I34. */
#define EC_FILE  103 /* File error of some sort. */
#define EC_SYS   104 /* Beyond our scope. */

/*---------------------------------------------------------------------------*/

/* Local external variables.
*/
static void (*DataOutFuncPtr)(unsigned thebyte) ;

static signed long ImageLen ; /* Number of bytes in device image portion of file. */
static unsigned short PortBase ; /* 4I34 I/O base address. (As set by jumpers.) */

static const char ProgNameMsg[] = { PROGNAME } ;

static FILE *ConfigFile ;

/*---------------------------------------------------------------------------*/

/* Function: signon
   Purpose: Display the signon message.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
*/

static void signon(void)

{
	printf("%s", SIGNONMESSAGE) ;
	printf("Mesa Electronics 4I34 configuration utility.\n") ;
}

/*---------------------------------------------------------------------------*/

/* Function: fatalerror
   Purpose: Display a fatal error message and exit.
   Used by: Local functions.
   Returns: Never.
   Notes:
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
	 2) Monday June 9, 2003 -- 15:21:41.
	    -Now beeps when displaying error message.
*/

static void fatalerror(int exitcode, const char *fmt, ...)

{
	va_list vp ;


	va_start(vp, fmt) ;
	fprintf(stderr, "\n\n\a%s:\n", ProgNameMsg) ;
	vfprintf(stderr, fmt, vp) ;
	putc('\n', stderr) ;

	exit(exitcode) ;
}

/*---------------------------------------------------------------------------*/

/* Function: readbytesfromfile
   Purpose: Read the specified number of bytes from the specified offset in
	 the specified open file.
   Used by: Local functions.
   Returns: The number of bytes read.
   Notes:
     -The file position is not restored to its original value.
     -Don't try to read 0 bytes.
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static size_t readbytesfromfile(FILE *f, long fromoffset, size_t numbytes, unsigned char *bufptr)

{
	if(fseek(f, fromoffset, SEEK_SET) != 0)
	{
		return 0 ;
	}
	return fread(bufptr, 1, numbytes, f) ;
}

/*---------------------------------------------------------------------------*/

/* Function: saybytesfromfile
   Purpose: Send the bytes from the specified offset in the specified open
     file to standard out.
   Used by: Local functions.
   Returns: The number of bytes read.
   Notes:
     -It is assumed that numbytes includes the trailing '\0'.
     -The file position is not restored to its original value.
     -Don't try to output 0 bytes.
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static size_t saybytesfromfile(FILE *f, long fromoffset, size_t numbytes)

{
	int c ;

	size_t len = 0 ;


	if(fseek(f, fromoffset, SEEK_SET) != 0)
	{
		return 0 ;
	}
	while(numbytes-- != 0)
	{
		if((c = getc(f)) == EOF)
		{
			break ;
		}
		++len ;
		if(c == '\0')
		{
			break ;
		}
		putchar(c) ;
	}
	return len ;
}

/*---------------------------------------------------------------------------*/

/* Function: saybytesfromfile
   Purpose: Send the bytes from the specified offset in the specified open
     file to standard out.
   Used by: Local functions.
   Returns: The number of bytes read.
   Notes:
     -It is assumed that numbytes includes the trailing '\0'.
     -The file position is not restored to its original value.
     -Don't try to output 0 bytes.
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static void sayfileinfo(signed long *infobaseptr)

{
	size_t len ;

	unsigned char b[2] ;


	if(readbytesfromfile(ConfigFile, *infobaseptr, sizeof(b), b) != sizeof(b))
	{
		ioerror:
		fatalerror(EC_FILE, "File I/O error.  Unexpected end of file?") ;
	}
	len = ((b[0] << 8) | b[1]) ; /* (Bigendian.) */
	if(saybytesfromfile(ConfigFile, (*infobaseptr + sizeof(b)), len) != len)
	{
		goto ioerror ;
	}
	*infobaseptr += (sizeof(b) + (signed long)len + 1) ; /* Skip over length field, text field, and tag. */
}

/*---------------------------------------------------------------------------*/

/* Function: outportbwswap
   Purpose: Send a byte to the 4I34 data port with bits swapped.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static void outportbwswap(unsigned thebyte)

{
	static unsigned char swaptab[256] =
	{
		0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
		0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
		0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
		0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
		0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
		0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
		0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
		0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
		0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
		0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
		0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
		0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
		0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
		0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
		0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
		0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
	} ;


	outportb((PortBase + R_4I34DATA), swaptab[thebyte]) ;
}

/*---------------------------------------------------------------------------*/

/* Function: outportbnoswap
   Purpose: Send a byte to the 4I34 data port with no bit swapping.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static void outportbnoswap(unsigned thebyte)

{
	outportb((PortBase + R_4I34DATA), thebyte) ;
}

/*---------------------------------------------------------------------------*/

/* Function: filelengthq
   Purpose: Determine the length of the specified file.
   Used by: Local functions.
   Returns: The file length, or -1 on failure.
   Notes:
	 -The file's location pointer is preserved.
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
*/

static long filelengthq(FILE *f)

{
	long curpos, len ;


	curpos = ftell(f) ;
	if(curpos < 0)
	{
		return -1 ;
	}
	if(fseek(f, 0, SEEK_END) != 0)
	{
		return -1 ;
	}
	len = ftell(f) ;
	if(len < 0)
	{
		return -1 ;
	}
	if(fseek(f, curpos, SEEK_SET) != 0)
	{
		return -1 ;
	}
	return len ;
}

/*---------------------------------------------------------------------------*/

/* Function: selectfiletype
   Purpose: Set up for processing the input file.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
     1) Monday June 9, 2003 -- 15:21:41.
*/

static void selectfiletype(void)

{
	unsigned char b[14] ;


	/* Read in the interesting parts of the file header.
	*/
	if(readbytesfromfile(ConfigFile, 0, sizeof(b), b) != sizeof(b))
	{
		ioerror:
		fatalerror(EC_FILE, "File I/O error.  Unexpected end of file?") ;
	}

	/* Figure out what kind of file we have.
	*/
	if((b[0] == 0x00) && (b[1] == 0x09) && (b[11] == 0x00) && (b[12] == 0x01) && (b[13] == 'a'))
	{	/* Looks like a .BIT file.
		*/
		signed long base ;


		printf("\nLooks like a .BIT file:") ;
		base = 14 ; /* Offset of design name length field. */

		/* Display file particulars.
		*/
		printf("\nDesign name:          ") ; sayfileinfo(&base) ;
		printf("\nPart I.D.:            ") ; sayfileinfo(&base) ;
		printf("\nDesign date:          ") ; sayfileinfo(&base) ;
		printf("\nDesign time:          ") ; sayfileinfo(&base) ;

		if(readbytesfromfile(ConfigFile, base, 4, b) != 4)
		{
			goto ioerror ;
		}
		ImageLen = (((unsigned long)b[0] << 24) |
					((unsigned long)b[1] << 16) |
					((unsigned long)b[2] << 8) |
					(unsigned long)b[3]) ;
		printf("\nConfiguration length: %lu bytes", ImageLen) ;

		DataOutFuncPtr = outportbwswap ; /* We have to swap bits in .BIT files. */

		/* We leave the file position set to the next byte in the file, which should
		     be the first byte of the body of the data image.
		*/
	}
	else if((b[0] == 0xFF) && (b[4] == 0x55) && (b[5] == 0x99) && (b[6] == 0xAA) && (b[7] == 0x66))
	{	/* Looks like a PROM file.
		*/
		printf("\nLooks like a PROM file:") ;
		DataOutFuncPtr = outportbnoswap ; /* No bit swap in PROM files. */

		ImageLen = filelengthq(ConfigFile); ;
		if(ImageLen < 0)
		{
			seekerror:
			fatalerror(EC_SYS, "File seek error.") ;
		}
		printf("\nConfiguration length: %lu bytes", ImageLen) ;

		/* PROM file data starts at offset 0, so we have to back up.
		*/
		if(fseek(ConfigFile, 0, SEEK_SET) != 0)
		{
			goto seekerror ;
		}
	}
	else
	{	/* It isn't something we know about.
		*/
		fatalerror(EC_FILE, "Unknown file type.") ;
	}

	putchar('\n') ;
	fflush(stdout) ; /* (Make sure the user can see all output to date.) */
}

/*---------------------------------------------------------------------------*/

/* Function: pcl
   Purpose: Collect parameters from the command line.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
     1) Thursday October 4, 2001 -- 18:08:08.
     2) Monday June 3, 2002 -- 15:36:51.
		-Now calls selectfiletype().
		-I/O addresses can now be anywhere in 0x0100...0x03FF.
*/

static const char FileNameInfo[] = "\n"
"filename.ext specifies the name of the configuration file to be loaded." ;

static const char PortInfo[] = "\n"
"port is the 3-digit hexadecimal I/O location of the card." ;

static const char PurposeInfo[] = "\n"
"Purpose: 4I34 configuration upload." ;


static void pcl(unsigned argc, char *argv[])

{
	if(argc != 3)
	{
		printf("\n%s program usage:\n%s filename.ext port", ProgNameMsg, ProgNameMsg) ;
		printf("\nWhere:") ;
		printf("%s", FileNameInfo) ;
		printf("%s", PortInfo) ;
		printf("%s", PurposeInfo) ;
		putchar('\n') ;

		exit(EC_BADCL) ;
	}

	if((ConfigFile = fopen(argv[1], "rb")) == 0) /* (Binary mode.) */
	{
		fatalerror(EC_USER, "File not found: %s", argv[1]) ;
	}
	selectfiletype() ;

	if(sscanf(argv[2], "%x", &PortBase) != 1)
	{
		badhex:
		fatalerror(EC_USER, "Invalid hexadecimal port address: %s", argv[2]) ;
	}
	if((PortBase < 0x0100) || (PortBase > 0x03FF) || ((PortBase & (NUM4I34REGS - 1)) != 0))
	{
		goto badhex ;
	}
}

/*---------------------------------------------------------------------------*/

/* Function: wait4dun
   Purpose: Wait for the card to finish programming.
   Used by: Local functions.
   Returns: Success/failure status.
   Notes:
	 -Failure to become DONE may be caused by an attempt to load an invalid
		FPGA code image, or an I/O address conflict.
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
*/

static signed wait4dun(void)

{
	unsigned waitcount ;


	/* Wait for the DONE bit to indicate completion of programming.
		 (First-revision 4I34 cards have no pullup on R_4I34STATUS.B_4I34PROGDUN,
		 so it may take a while for the bit to got to its "done" state.
	*/
	for(waitcount = PROGWAITLOOPCOUNT ; waitcount != 0 ; --waitcount)
	{
		if(inportb(PortBase + R_4I34STATUS) & M_4I34PROGDUN)
		{	/* Programming seems to be complete; make sure the state sticks and we're not
				 seeing some sort of I/O conflict-related foo.
			*/
			for(waitcount = 100 ; waitcount != 0 ; --waitcount)
			{
				if(!inportb(PortBase + R_4I34STATUS) & M_4I34PROGDUN)
				{
					return -1 ; /* Shouldn't. */
				}
			}

			return 0 ; /* Success. */
		}
	}

	return -1 ; /* Failure -- didn't become DONE. */
}

/*---------------------------------------------------------------------------*/

/* Function: programcard
   Purpose: Initialize the 4I34 card at port PortBase with the contents of
	 file ConfigFile.
   Used by: Local functions.
   Returns: Nothing.
   Notes:
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
	 2) Monday June 9, 2003 -- 15:21:41.
	    -Data is now sent to the card via *DataOutFuncPtr instead of directly
		   calling outportb().
*/

#define FORMATSTRING "%-6lu"

static void programcard(void)

{
	signed long bytesleft ;
	unsigned long numbytes ;

	size_t count ;

	static unsigned char filebuffer[8192] ;
	unsigned char *bptr ;


	bytesleft = ImageLen ;

#if defined AS_LINUX
	/* Enable direct access to the I/O port range.
	*/
	if(ioperm((unsigned long)PortBase, NUM4I34REGS, !0) < 0)
	{
		fatalerror(EC_SYS, "Can't enable direct port access.") ;
	}
#endif

	/* Enable programming.
	*/
	printf("\nProgramming...\n") ;
	fflush(stdout) ; /* (Make sure the user can see all output to date.) */

	outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSOFF | M_4I34CFGINITASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDOFF)) ;
	if(inportb(PortBase + R_4I34STATUS) & M_4I34PROGDUN)
	{	/* Note that if we see DONE at the start of programming, it's most likely due
			 to an attempt to access the 4I34 at the wrong I/O location.
		*/
		outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDOFF)) ;

		fatalerror(EC_HDW,
				   "<DONE> status bit indicates busy at start of programming.  I/O address\n"
				   "  conflict?") ;
	}
	outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSON | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEENABLE | M_4I34LEDON)) ;

	/* Delay for at least 100 uS. to allow the FPGA to finish its reset
		 sequencing.  (In reality, the time taken for the initial file access
		 makes this this delay redundant.)
	*/
	for(count = 1000 ; count != 0 ; --count)
	{
		inportb(PortBase + R_4I34STATUS) ;
	}

	/* Program the card.
	*/
	numbytes = 0 ;
	for( ; ; )
	{	/* Write the file to the 4I34.
		*/
		errno = 0 ;
		count = fread(filebuffer, 1, sizeof(filebuffer), ConfigFile) ;
		if(count == 0)
		{	/* 0 bytes could be end of file or error; must check.
			*/
			if(errno != 0)
			{
				outportb((PortBase + R_4I34CONTROL),
					(M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDOFF)) ;

				fatalerror(EC_SYS, "File I/O error.") ;
			}
			else
			{
				printf(FORMATSTRING, 0L) ;

				outportb((PortBase + R_4I34CONTROL),
					(M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDON)) ;
				fclose(ConfigFile) ;

				break ; /* Done. */
			}
		}
		else
		{	/* Write the block to the 4I34 data port.
			*/
			printf(FORMATSTRING "\r", bytesleft) ;
			fflush(stdout) ;
			numbytes += (unsigned long)count ;
			bytesleft -= (signed long)count ;

			bptr = filebuffer ;
			do
			{
				DataOutFuncPtr(*bptr++) ;
			} while(--count != 0) ;
		}
	}

	/* Wait for completion of programming.
	*/
	outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDON)) ;
	if(wait4dun() < 0)
	{
		outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDOFF)) ;

		fatalerror(EC_HDW,
				   "Error: <DONE> status bit not set; programming not completed. I/O address\n"
				   "  conflict?") ;
	}

	/* Indicate end of programming.
	*/
	outportb((PortBase + R_4I34CONTROL), (M_4I34CFGCSOFF | M_4I34CFGINITDEASSERT | M_4I34CFGWRITEDISABLE | M_4I34LEDOFF)) ;

	/* Send configuration completion clocks.  (6 should be enough, but we send
		 lots for good measure.)
	*/
	for(count = 24 ; count != 0 ; --count)
	{
		outportb((PortBase + R_4I34DATA), 0xFF) ;
	}

	printf("\nSuccessfully programmed %lu bytes.", numbytes) ;
}

/*---------------------------------------------------------------------------*/

/* Function: main
   Purpose: Program entry point.
   Used by: Program launcher.
   Returns: Exit status.
   Notes:
   Revision history:
	 1) Thursday October 4, 2001 -- 18:08:08.
*/

int main(int argc, char *argv[])

{
	signon() ;
	pcl(argc, argv) ;
	programcard() ;
	printf("\n\nFunction complete.\n") ;

	return EC_OK ;
}

/*---------------------------------------------------------------------------*/
